unit MainForm;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ColorPalette, TurboSprite, DestinationSpriteMover, PolygonSprite, Polygon,
  StdCtrls, FireworkSprite, KeyboardHandler, ThrustSpriteMover, Buttons,
  Missile, DibCanvas, ComCtrls;

type
  TForm1 = class(TForm)
    surface: TSpriteSurface;
    pal: TColorPalette;
    engineAsteroids: TSpriteEngine;
    moverDest: TDestinationSpriteMover;
    psfAstMedium: TPolygonSpriteFactory;
    psfAstSmall: TPolygonSpriteFactory;
    psfAstLarge: TPolygonSpriteFactory;
    fwsf: TFireworkSpriteFactory;
    keyboard: TKeyboardHandler;
    psfShip: TPolygonSpriteFactory;
    engineShip: TSpriteEngine;
    moverThrust: TThrustSpriteMover;
    btnStart: TSpeedButton;
    engineMissiles: TSpriteEngine;
    engineExplosions: TSpriteEngine;
    Label1: TLabel;
    lblShots: TStaticText;
    Label2: TLabel;
    lblHits: TStaticText;
    Label3: TLabel;
    Label4: TLabel;
    lblGrossScore: TStaticText;
    lblNetScore: TStaticText;
    Label5: TLabel;
    lblMarks: TStaticText;
    psfFire: TPolygonSpriteFactory;
    Label6: TLabel;
    lblLives: TStaticText;
    fsfShip: TFireworkSpriteFactory;
    StatusBar1: TStatusBar;
    procedure Button1Click(Sender: TObject);
    procedure moverDestInitSprite(Sender: TObject; ASprite: TSprite);
    procedure engineAsteroidsRemoveSprite(Sender: TObject;
      ASprite: TSprite);
    procedure moverThrustInitSprite(Sender: TObject;
      ASprite: TSprite);
    procedure surfaceBeforeRender(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure engineMissilesCollision(Sender: TObject; Sprite1,
      Sprite2: TSprite);
    procedure surfaceAfterRender(Sender: TObject);
    procedure engineShipCollision(Sender: TObject; Sprite1,
      Sprite2: TSprite);
  private
    bUpdate: boolean;
    procedure SetupWave;
    procedure UpdateScores;
  public
    GameInProgress: boolean;
    MissileLaunched: boolean;
    ShotsFired: integer;
    AsteroidCount: integer;
    HitCount: integer;
    Score: integer;
    Invulnerable: integer;
    BlinkCount: integer;
    Lives: integer;
    WaitingForWave: integer;
    WaitingAfterDead: integer;
    Hyperspace: integer;
    FreeShip: integer;
    procedure GameOver; virtual;
  end;

var
  Form1: TForm1;

implementation

{$R *.DFM}

procedure TForm1.Button1Click(Sender: TObject);
var
  s: TSprite;
begin
  Randomize;
  GameInProgress := false;
  surface.Active := true;
  engineAsteroids.RemoveAllSprites;
  engineMissiles.RemoveAllSprites;
  engineShip.RemoveAllSprites;
  engineShip.AddSprite( psfShip.CreateSprite );
  s := psfFire.CreateSprite;
  engineShip.AddSprite( s );
  s.Visible := false;
  s.Tag := 10;
  s.Priority := 1;
  AsteroidCount := 3;
  HitCount := 0;
  ShotsFired := 0;
  Score := 0;
  Lives := 3;
  FreeShip := 5000;
  SetupWave;
  GameInProgress := true;
  btnStart.Visible := false;
  UpdateScores;
end;

procedure TForm1.moverDestInitSprite(Sender: TObject; ASprite: TSprite);
begin
  if Random > 0.5 then
    ASprite.SpinDirection := dirBackward;
  if Random > 0.4 then
    ASprite.SpinSpeed := Random * Random * 20;
end;

procedure TForm1.engineAsteroidsRemoveSprite(Sender: TObject;
  ASprite: TSprite);
var
  ps, psNew: TSprite;
  fws: TSprite;
  procedure createChunks( sf: TSpriteFactory; nTag: integer );
  var
    i: integer;
  begin
    for i := 0 to Random( 3 ) + 1 do
    begin
      psNew := sf.CreateSprite;
      psNew.Tag := nTag;
      engineAsteroids.AddSprite( psNew );
      psNew.PositionX := ps.PositionX;
      psNew.PositionY := ps.PositionY;
    end;
  end;
begin
  if GameInProgress then
  begin
    ps := ASprite;
    fws := fwsf.CreateSprite;
    engineExplosions.AddSprite( fws );
    fws.PositionX := ps.PositionX;
    fws.PositionY := ps.PositionY;
    moverDest.Speed[fws] := 0;
    if ps.Tag = 2 then
      createChunks( psfAstMedium, 1 )
    else if ps.Tag = 1 then
      createChunks( psfAstSmall, 0 );
  end;
end;

procedure TForm1.moverThrustInitSprite(Sender: TObject;
  ASprite: TSprite);
begin
  ASprite.PositionX := surface.VirtualWidth div 2;
  ASprite.PositionY := surface.VirtualHeight div 2;
end;

(* This event gets called before each animation cycle *)
procedure TForm1.surfaceBeforeRender(Sender: TObject);
var
  s: TPolygonSprite;
  m, f: TSprite;
  x, y: integer;
begin
(* If the ship is created, do processing for it *)
  if engineShip.SpriteCount > 1 then
  begin
    s := engineShip.Sprites[0] as TPolygonSprite;
    f := engineShip.Sprites[1];
(* If this is a pause after the ship was killed, do nothing *)
    if WaitingAfterDead > 0 then
    begin
      f.Visible := false;
      Dec( WaitingAfterDead );
      if WaitingAfterDead = 0 then
      begin
        if Lives = 0 then
        begin
          GameOver;
          Exit;
        end;
        s.Visible := true;
        Invulnerable := 100;
        s.PositionX := surface.VirtualWidth div 2;
        s.PositionY := surface.VirtualHeight div 2;
        f.PositionX := surface.VirtualWidth div 2;
        f.PositionY := surface.VirtualHeight div 2;
        moverThrust.SpeedX[s] := 0;
        moverThrust.SpeedY[s] := 0;
        moverThrust.SpeedX[f] := 0;
        moverThrust.SpeedY[f] := 0;
        s.Angle := 0;
        f.Angle := 0;
      end;
      Exit;
    end;
(* If the ship is in hyperspace do nothing *)
    if Hyperspace > 0 then
    begin
      Dec( Hyperspace );
      if Hyperspace = 0 then
      begin
        s.Visible := true;
        s.RandomizePosition;
        f.PositionX := s.PositionX;
        f.PositionY := s.PositionY;
        moverThrust.SpeedX[s] := 0;
        moverThrust.SpeedY[s] := 0;
        moverThrust.SpeedX[f] := 0;
        moverThrust.SpeedY[f] := 0;
      end;
      Exit;
    end;
(* Poll keyboard for hyperspace *)
    if keyboard.VirtualKeyDown[VK_ESCAPE] and ( Hyperspace = 0 ) then
    begin
      Hyperspace := 100;
      s.Visible := false;
      f.Visible := false;
    end;
(* Poll the keyboard and rotate the ship *)
    if keyboard.KeyDown['k'] then
      s.Angle := s.Angle - 10
    else if keyboard.KeyDown['l'] then
      s.Angle := s.Angle + 10;
    f.Angle := s.Angle;
(* Apply thrust if spacebar is pressed *)
    if keyboard.KeyDown[' '] then
      moverThrust.Thrust[s] := 0.3
    else
      moverThrust.Thrust[s] := 0;
(* Need to apply the same thrust to the ship fire sprite *)
    moverThrust.Thrust[f] := moverThrust.Thrust[s];
(* Ship fire blinks in and out randomly *)
    f.Visible := ( moverThrust.Thrust[f] > 0 ) and ( Random > 0.4 ) and s.Visible;
(* See if the enter key is still down from launching a missile *)
    if MissileLaunched then
    begin
      if not keyboard.VirtualKeyDown[VK_RETURN] then
        MissileLaunched := false;
    end
    else
    begin
(* Only six missiles at a time! *)
      if engineMissiles.SpriteCount > 5 then
        Exit;
(* Launch a missile if they press enter *)
      if keyboard.VirtualKeyDown[13] then
      begin
        m := TMissile.Create;
        Inc( ShotsFired );
        bUpdate := true;
        engineMissiles.AddSprite( m );
(* Missile position equals ship position *)
        m.PositionX := s.PositionX + s.Polygon.Vertices[0].X;
        m.PositionY := s.PositionY + s.Polygon.Vertices[0].Y;
(* Determine the missile destination by projecting the current
   trajectory of the ship *)
        x := Trunc( s.PositionX + 100 * Cos( ( 90 - s.Angle ) * PI / 180 ) );
        y := Trunc( s.PositionY - 100 * Sin( ( 90 - s.Angle ) * PI / 180 ) );
(* Set the speed and destination of the missile *)
        moverDest.Speed[m] := 5;
        moverDest.Destination[m] := Point( x, y );
(* Adjust the speed of the missile based on the speed of the ship *)
        moverDest.SpeedX[m] := moverDest.SpeedX[m] + moverThrust.SpeedX[s];
        moverDest.SpeedY[m] := moverDest.SpeedY[m] + moverThrust.SpeedY[s];
        MissileLaunched := true;
      end;
    end;
(* The ship blinks when it is invulnerable, for a few seconds
   after it is created *)
    if Invulnerable > 0 then
    begin
      Dec( Invulnerable );
      Inc( BlinkCount );
      if BlinkCount > 3 then
      begin
        s.Visible := not s.Visible;
        BlinkCount := 0;
      end;
    end
    else
      s.Visible := Hyperspace = 0;
(* We might be waiting before setting up a new wave *)
    if WaitingForWave > 0 then
    begin
      Dec( WaitingForWave );
      if WaitingForWave = 0 then
      begin
        SetupWave;
      end;
      Exit;
    end;
(* If all the asteroids are destroyed, begin a new wave *)
    if engineAsteroids.SpriteCount = 0 then
    begin
      WaitingForWave := 50;
      Exit;
    end;
  end;
end;

procedure TForm1.FormCreate(Sender: TObject);
begin
  engineMissiles.AddCollisionEngine( engineAsteroids );
  engineShip.AddCollisionEngine( engineAsteroids );
  engineShip.AddCollisionEngine( engineAsteroids );
end;

procedure TForm1.engineMissilesCollision(Sender: TObject; Sprite1,
  Sprite2: TSprite);
begin
  if not GameInProgress then
    Exit;
  if Sprite2.Tag = 10 then
    Exit;
  Sprite1.Dead := true;
  Sprite2.Dead := true;
  Inc( HitCount );
  case Sprite2.Tag of
    0:
      Inc( Score, 20 );
    1:
      Inc( Score, 15 );
    2:
      Inc( Score, 10 );
  end;
  if Score > FreeShip then
  begin
    FreeShip := FreeShip + 5000;
    Inc( Lives );
    lblLives.Caption := IntToStr( Lives );
  end;
  bUpdate := true;
end;

procedure TForm1.SetupWave;
var
  s: TSprite;
  i: integer;
begin
  Invulnerable := 100;
  for i := 1 to AsteroidCount do
  begin
    s := psfAstLarge.CreateSprite;
    s.Tag := 2;
    engineAsteroids.AddSprite( s );
  end;
  Inc( AsteroidCount );
end;

procedure TForm1.surfaceAfterRender(Sender: TObject);
begin
  if bUpdate then
  begin
    bUpdate := false;
    UpdateScores;
  end;
end;

procedure TForm1.UpdateScores;
begin
  lblShots.Caption := IntToStr( ShotsFired );
  lblHits.Caption := IntToStr( HitCount );
  lblGrossScore.Caption := FOrmatFloat( '#,##0', Score );
  if ShotsFired > 0 then
  begin
    lblNetScore.Caption := FormatFloat( '#,##0', Trunc( Score * ( HitCount / ShotsFired ) ) );
    lblMarks.Caption := FormatFloat( '##0.00', HitCount / ShotsFired * 100 ) + '%';
  end
  else
  begin
    lblNetScore.Caption := '0';
    lblMarks.Caption := '0';
  end;
  lblLives.Caption := IntToStr( Lives );
end;

procedure TForm1.engineShipCollision(Sender: TObject; Sprite1,
  Sprite2: TSprite);
var
  exp: TSprite;
begin
  if not GameInProgress then
    Exit;
  if Sprite1.Tag = 0 then
    if Invulnerable = 0 then
    begin
      Sprite1.Visible := false;
      engineAsteroids.RemoveSprite( Sprite2 );
      exp := fsfShip.CreateSprite;
      exp.PositionX := Sprite1.PositionX;
      exp.PositionY := Sprite1.PositionY;
      engineExplosions.AddSprite( exp );
      WaitingAfterDead := 100;
      Dec( Lives );
      lblLives.Caption := IntToStr( Lives );
    end;
end;

procedure TForm1.GameOver;
begin
  GameInProgress := false;
  engineAsteroids.RemoveAllSprites;
  engineShip.RemoveAllSprites;
  showMessage( 'G A M E   O V E R' );
  btnStart.Visible := true;
end;

end.
